﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [Item("Z-Score Normalization", "Z-Score normalization transformation to standardize the values to a Target Mean and Target Standard Deviation.")]
  [StorableClass]
  public class ZNormalizationTransformation : Transformation<double> {
    #region Parameters
    private IFixedValueParameter<DoubleValue> TargetMeanParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["Target Mean"]; }
    }
    private IFixedValueParameter<DoubleValue> TargetStandardDeviationParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["Target Standard Deviation"]; }
    }
    private IFixedValueParameter<DoubleValue> OriginalMeanParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["Original Mean"]; }
    }
    private IFixedValueParameter<DoubleValue> OriginalStandardDeviationParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["Original Standard Deviation"]; }
    }
    #endregion

    #region Properties
    public double TargetMean {
      get { return TargetMeanParameter.Value.Value; }
      set { TargetMeanParameter.Value.Value = value; }
    }
    public double TargetStandardDeviation {
      get { return TargetStandardDeviationParameter.Value.Value; }
      set { TargetStandardDeviationParameter.Value.Value = value; }
    }
    public double OriginalMean {
      get { return OriginalMeanParameter.Value.Value; }
      private set { OriginalMeanParameter.Value.Value = value; }
    }
    public double OriginalStandardDeviation {
      get { return OriginalStandardDeviationParameter.Value.Value; }
      private set { OriginalStandardDeviationParameter.Value.Value = value; }
    }
    #endregion

    #region Constructor, Cloning & Persistence
    public ZNormalizationTransformation()
      : base() {
      Parameters.Add(new FixedValueParameter<DoubleValue>("Target Mean", new DoubleValue(0)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("Target Standard Deviation", new DoubleValue(1)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("Original Mean", new DoubleValue(double.NaN)) { Hidden = true });
      Parameters.Add(new FixedValueParameter<DoubleValue>("Original Standard Deviation", new DoubleValue(double.NaN)) { Hidden = true });
    }

    protected ZNormalizationTransformation(ZNormalizationTransformation original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ZNormalizationTransformation(this, cloner);
    }

    [StorableConstructor]
    protected ZNormalizationTransformation(bool deserializing)
      : base(deserializing) {
    }
    #endregion

    public override bool Check(IEnumerable<double> data, out string errorMessage) {
      if (data.StandardDeviationPop().IsAlmost(0.0)) {
        errorMessage = "Z-Score Normalization cannot be applied for data with a standard deviation of zero.";
        return false;
      }
      return base.Check(data, out errorMessage);
    }

    public override void Configure(IEnumerable<double> data) {
      OriginalMean = data.Average();
      OriginalStandardDeviation = data.StandardDeviationPop();
      base.Configure(data);
    }

    public override IEnumerable<double> Apply(IEnumerable<double> data) {
      if (double.IsNaN(OriginalMean) || double.IsNaN(OriginalStandardDeviation)) //TODO isConfigured field?
        Configure(data);

      return Apply(data, TargetMean, TargetStandardDeviation, OriginalMean, OriginalStandardDeviation);
    }

    public override IEnumerable<double> InverseApply(IEnumerable<double> data) {
      return InverseApply(data, TargetMean, TargetStandardDeviation, OriginalMean, OriginalStandardDeviation);
    }


    public static IEnumerable<double> Apply(IEnumerable<double> data, double targetMean, double targetStandardDeviation, double originalMean = double.NaN, double originalStandardDeviation = double.NaN) {
      if (double.IsNaN(originalMean)) originalMean = data.Average();
      if (double.IsNaN(originalStandardDeviation)) originalStandardDeviation = data.StandardDeviationPop();

      return data
        .Select(x => (x - originalMean) / originalStandardDeviation) // standardize (0, 1)
        .Select(x => x * targetStandardDeviation + targetMean);
    }

    public static IEnumerable<double> InverseApply(IEnumerable<double> data, double targetMean, double targetStandardDeviation, double originalMean, double originalStandardDeviation) {
      return Apply(data, originalMean, originalStandardDeviation, targetMean, targetStandardDeviation);
    }
  }
}